% Bereken 'n impulsrespons (as 32 tapgewigte) vir die G3RUH-modem met inagname van die
% frekwensiegedrag van die in- en uittreefilters en kanaaleienskappe soos bepaal deur die
% FM-sender/ontvanger.  Sodoende word die kanaal ge-effen om 'n onvervormde oogpatroon aan die
% ontvangkant te lewer.

% Johann Lochner (8/11/97)

% Die volgende ler word geskep:
% tapgewig.mat	32 tapgewigte vir transversale filter

clear all;

% ------------------------------ Opstelling van veranderlikes -----------------------------------

H_kanaal=[1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 ...		% spesifiseer die frekwensiegedrag van die kanaal tussen 0 en 9600 Hz met intervalle van 300 Hz
		1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1];
tapafset=4;						% skuif tapgewigte sodat beide endwaardes so klein as moontlik is: sodoende word die die effek van venstering geminimeer

bisse=8;						% aantal bisse beskou vir pulsvorming
fb=9600;						% bisfrekwensie
TX_fs=4*fb;						% sender-monsterfrekwensie

alfa=0.5;						% vir gehefde cosinus-spektrum

load filters;						% laai koffisinte van TX-filter en RX-filter

% ----------------------- Bereken die ideale gehefde cosinus-spektrum ---------------------------

f=0:fb/(bisse*TX_fs/fb):TX_fs/2;			% frekwensie vanaf nul tot by Nyquist
f1=(1-alfa)*fb/2;

for i=1:length(f)					% gehefde cosinus
  if f(i) < f1
    ideale_spektrum(i)=1;
  elseif f(i) < fb-f1
    ideale_spektrum(i)=1/2*(1+cos(pi*(f(i)-f1)/(fb-2*f1)));
  else
    ideale_spektrum(i)=0;
  end;
end;

neg_spektrum=conj(fliplr(ideale_spektrum(2:length(ideale_spektrum))));
ideale_spektrum=[ideale_spektrum neg_spektrum];		% ideale spektrum vanaf gs tot by die monsterfrekwensie

% ------------------- Bereken die frekwensiegedrag van die S/A-omsetter -------------------------

H_zoh=sinc(f/TX_fs).*exp(-j*pi*f/TX_fs);
neg_spektrum=conj(fliplr(H_zoh(2:length(H_zoh))));
H_zoh=[H_zoh neg_spektrum];

% ------------- Bereken die frekwensiegedrag van die anti-oorvleuelingsfilter -------------------

TX_spektrum=freqs(TX_teller,TX_noemer,2*pi*f); TX_spektrum(1)=TX_spektrum(2);
neg_spektrum=conj(fliplr(TX_spektrum(2:length(TX_spektrum))));
TX_spektrum=[TX_spektrum neg_spektrum];

% ----------------------- Bereken die frekwensiegedrag van die kanaal ---------------------------

H_kanaal=[H_kanaal ones(1,length(f)-length(H_kanaal))];
neg_spektrum=conj(fliplr(H_kanaal(2:length(H_kanaal))));
H_kanaal=[H_kanaal neg_spektrum];

% ------------------ Bereken die frekwensiegedrag van die ontvangerfilter -----------------------

RX_spektrum=freqs(RX_teller,RX_noemer,2*pi*f); RX_spektrum(1)=RX_spektrum(2);
neg_spektrum=conj(fliplr(RX_spektrum(2:length(RX_spektrum))));
RX_spektrum=[RX_spektrum neg_spektrum];

% --------------------------------- Bereken die tapgewigte --------------------------------------

H_tappe=ideale_spektrum./(RX_spektrum.*H_kanaal.*TX_spektrum.*H_zoh);	% bereken die frekwensie-spektrum van die impulsrespons

tappe=-real(ifft(H_tappe));				% bereken die tapgewigte deur die inverse DFT
tappe=[tappe(length(tappe)-TX_fs*bisse/(2*fb)+1-tapafset:length(tappe)) tappe(1:TX_fs*bisse/(2*fb)-tapafset)];	% venster 32 tappe m.b.v. 'n reghoekvenster
%tappe=tappe.*hamming(length(tappe))';			% as alternatief vir die reghoekvenster
tappe=TX_fs/fb*tappe;					% bring skalering a.g.v. zoh in berekening

save tapgewig tappe					% stoor tapgewigte

% ------------------------------------------ Afvoer ---------------------------------------------

figure(1);						% vertoon aanwins- en faseweergawe van die ge-effende kanaal
clf;

subplot(2,1,1);
plot(20*log10(ideale_spektrum));
hold on;
plot(20*log10(abs(RX_spektrum.*H_kanaal.*TX_spektrum.*H_zoh.*fft(tappe,length(H_kanaal)))*fb/TX_fs),'c');
plot(20*log10(abs(RX_spektrum.*H_kanaal.*TX_spektrum.*H_zoh)),'r');
asse=axis;
axis([asse(1:2) -30 10]);

subplot(2,1,2);
plot(phase(RX_spektrum.*H_kanaal.*TX_spektrum.*H_zoh.*fft(tappe,length(H_kanaal))),'c');
hold on
plot(phase(RX_spektrum.*H_kanaal.*TX_spektrum.*H_zoh),'r');
zoom on;

figure(2);						% vertoon die berekende tapgewigte
clf;
stem(tappe);
zoom on;