% Simulasie van die G3RUH-radiomodem.

% Johann Lochner (1 Maart 1999)

% Die volgende lers word gebruik:
% odf.m		genereer dataler filters.mat (nie nodig wanneer die simulasie loop)
% filters.mat	filterkoffisinte van TX-filter en RX-filter
% effen.m	genereer dataler tapgewig.mat (nie nodig wanneer die simulasie loop)
% tapgewig.mat	32 tapgewigte vir transversale filter
% eprom.bin	G3RUH-modem se EPROM-data
% des2bin.m	skakel desimale golfvormkeuse om na 'n binre adres vir die EPROM
% lromdata.m	bepaal tapgewigte volgens EPROM-data
% tex2bin.m	skakel teksboodskap om na binre string (vr modulasie)
% bin2tex.m	skakel binre string om na teksboodskap (n demodulasie)
% spek.m	bereken spektrale inhoud van syferry

clear all;
begintyd=cputime;

% ===================================== Simulasie-parameters =====================================

pulsnommer=16;						% 0 tot 15: keuse van ROM-golfvorms; hor as 15: wek golfvorm op volgens spesifikasie

bisse=8;						% aantal bisse beskou vir pulsvorming
fb=9600;						% bisfrekwensie
TX_fs=4*fb;						% sender-monsterfrekwensie (kan hor gestel word wanneer golfvorm self opgewek word)
RX_fs=128*fb;						% ontvanger-monsterfrekwensie

TX_opstel=[1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0];		% moet 16 bisse lank wees vir die skommelaar
merker='##';						% dui die begin van die boodskap aan
boodskap='Toets';					% word in binre vorm oor die modem gestuur

alfa=0.5;						% vir gehefde cosinus-spektrum
SNR=15;							% sein tot ruis-verhouding in dB
oge=2;							% aantal o in oogkarakteristiek

load filters;						% laai koffisinte van TX-filter en RX-filter

% --------------------------- Bepaal die modem se transversale filter ---------------------------

load tapgewig						% gebruik tapgewigte soos bepaal deur effen.m

if pulsnommer < 16					% behalwe as 'n impulsrespons in die EPROM gekies is
  tappe=lromdata('eprom.bin',pulsnommer);
end;

% ------------------------------------ Demodulator-veranderlikes --------------------------------

kloksiklus=[ones(1,RX_fs/(2*fb)) zeros(1,RX_fs/(2*fb))];% vektor met enkele digitale kloksiklus
wyser=38;						% dui posisie van RX_klok in kloksiklus aan
ideale_wyser=65;					% dui posisie van ideale herwinningsklok in kloksiklus aan
flank=0;						% tel die afgaande flanke van RX_klok

RC=100e3*100e-9;					% RC-tyd van DCD-LDF
drempel=47e3/(47e3+33e3);				% waarbo sluit_rekord impliseer dat RX-klok en RX-proto in fase is

% =========================================== Modulasie =========================================

TX_data=tex2bin([merker boodskap]);			% stel merker en boodskap in binre vorm voor

TX_skommel=TX_opstel;					% skommel binre voorstelling
for i=1:length(TX_data);
  TX_skommel=[TX_skommel xor(TX_data(i), xor(TX_skommel(i+5),TX_skommel(i)))];
end;

fs_monster=[1; zeros(TX_fs/fb-1, 1)];			% monster die geskommelde ry teen TX-fs
TX_monster=fs_monster*TX_skommel;
TX_monster=TX_monster(:)';

TX_mod=conv(tappe,TX_monster)-0.5;			% moduleer die gemonsterde ry en verwyder GS-afset

fs_verskil=ones(RX_fs/TX_fs, 1);			% monster die gemoduleerde ry teen RX-fs
TX_zoh=fs_verskil*TX_mod;
TX_zoh=TX_zoh(:)';

t=linspace(0,length(TX_zoh)/RX_fs,length(TX_zoh));
TX_oudio=-lsim(TX_teller, TX_noemer, TX_zoh,t)';	% filter die trapsein by die zoh-uittree

% =========================================== Demodulasie =======================================

randn('seed',0);					% tel kanaalruis by
RX_oudio=TX_oudio+std(TX_oudio)*10^(-SNR/20)*randn(1,length(TX_oudio));
RX_oog=lsim(RX_teller, RX_noemer, RX_oudio,t)';		% filter die ontvangde sein om ruis te beperk
RX_dig=round((sign(RX_oog)+1)/2);			% beperk die gefilterde sein tot diskrete waardes

RX_proto=zeros(1,length(RX_dig));			% reserveer geheueruimte vir veranderlikes
RX_klok=zeros(1,length(RX_dig));
Ideale_klok=zeros(1,length(RX_dig));
sluit_rekord=0.25*ones(1,length(RX_dig));		% rekord van sluiting tussen tussen RX-klok en RX-proto
DCD_rekord=zeros(1,length(RX_dig)*fb/RX_fs);		% rekord van DCD-toestand by monsteroomblikke
afset_rekord=zeros(1,length(RX_dig)*fb/RX_fs);		% rekord van speling tussen RX-klok en ideale herwinningsklok by afgaande flanke van RX-klok
RX_skommel=zeros(1,length(RX_dig)*fb/RX_fs);		% rekord van ontvangde data (geskommel)

clc;

for i=1:length(RX_dig),					% word nie as matrikse bereken nie om die intydse situasie na te boots

  if i > RX_fs/(2*fb)					% onttrek die protoklok
    RX_proto(i)=xor(RX_dig(i), RX_dig(i-RX_fs/(2*fb)));
  else RX_proto(i)=0;
  end;

  if i > 1,
    if (RX_proto(i-1)==1) & (RX_proto(i)==0)		% by die afgaande flanke van die protoklok word RX-klok daarmee gesinchroniseer
      if wyser<RX_fs/(2*fb)+1				% as RX-klok hoog is
        wyser=wyser+1;					% versnel klok met een monsterpunt
      elseif wyser>RX_fs/(2*fb)+1			% as RX-klok laag is en nie die vorige monsteroomblik hoog was nie
        wyser=wyser-1;					% vertraag klok met een monsterpunt
      end;
    end;
  end;

  RX_klok(i)=kloksiklus(wyser);				% konstrueer RX-klok
  if wyser==length(kloksiklus)				% indeks na die golfvorm van RX-klok het maksimum bereik
    wyser=1;						% herstel na minimum
  else wyser=wyser+1;					% andersins loop die indeks aan
  end;

  Ideale_klok(i)=kloksiklus(ideale_wyser);		% doen dieselfde vir Ideale_klok
  if ideale_wyser==length(kloksiklus)
    ideale_wyser=1;

    home;
    disp([num2str(fix(i*fb/RX_fs)) ' van ' num2str(fix(length(RX_oog)*fb/RX_fs)) ' bisse gedemoduleer']);

  else ideale_wyser=ideale_wyser+1;
  end;

  if i>1						% laagdeurlaat-weergawe van ~RX-klok en RX-proto gee aanduiding van faseverskil tussen hierdie twee
    sluit_rekord(i)=exp(-1/(RC*RX_fs))*sluit_rekord(i-1)+(1-exp(-1/(RC*RX_fs)))*xor(RX_proto(i), ~RX_klok(i));

    if (RX_klok(i-1)==1) & (RX_klok(i)==0)		% by die afgaande flanke van RX-klok
      flank=flank+1;
      RX_skommel(flank)=RX_dig(i-1);			% word die geskommelde ry bepaal deur die digitale sein te monster

      if sluit_rekord(i)<drempel			% wanneer die speling tussen RX-klok en RX-proto se afgaande flanke groter is as toegelaat
        DCD=0;						% word DCD=0: draer verloor
      else DCD=1;					% of DCD=1: draer gevind
      end;

      DCD_rekord(flank)=DCD;				% rekord van DCD by monsteroomblikke
      afset_rekord(flank)=wyser-ideale_wyser;		% rekord van die speling tussen RX-klok en Ideale-klok by monsteroomblikke
    end;

  end;

end;

RX_skommel=[RX_skommel zeros(1,16)];			% sorteer die geskommelde data
for i=1:length(RX_skommel)-17
  RX_data(i)=xor(xor(RX_skommel(i+16), RX_skommel(i+5)), RX_skommel(i));
end;

% ============================================= Afvoer ==========================================

% Die res van die kode is oorwegend gemik op die vertoon van berekende data.  Standaard
% figuuruitleg-instruksies word gebruik en word gevolglik nie in detail gemotifeer nie.
% Dataverwerking-instruksies word steeds verantwoord.

min_f=0;						% vir skalering van asse in figure
max_f=TX_fs*1.3;
min_dB=-40;
max_dB=10;

% --------------- Vertoon die transversale filter se frekwensiegedrag (fig 1) -------------------

f=linspace(min_f,max_f,512);				% bereken die ideale gehefde cosinus-spektrum
f1=(1-alfa)*fb/2;
for i=1:length(f)
  if f(i) < f1
    ideale_spektrum(i)=1;
  elseif f(i) < fb-f1
    ideale_spektrum(i)=1/2*(1+cos(pi*(f(i)-f1)/(fb-2*f1)));
  else
    ideale_spektrum(i)=0;
  end;
end;

impuls_zoh=fs_verskil*tappe;				% bereken die impulsrespons van die transversale filter
impuls=impuls_zoh(:)';

spektrum=20*log10(abs(spek(impuls,2*pi*f/RX_fs))*fb/RX_fs);
fasehoek=unwrap(angle(spek(impuls,2*pi*f/RX_fs)))*180/pi;

figure(1);						
clf;

subplot(2,1,1);
hold on;

plot(f,spektrum,'c');

if alfa == 0						% vir die grensgeval van alfa=0 slegs reguit lyne
  line([min_f fb/2], [0 0]);
  line([fb/2 fb/2], [0 -100]);
else
  plot(f,20*log10(ideale_spektrum), 'y:');
end;

hold off;
axis([min_f max_f min_dB max_dB]);
title(['Aanwins- en faseweergawe van tranversale filter (geel: alfa=' sprintf('%3.2f',alfa) ')']);
ylabel('Aanwins [dB]');

subplot(2,1,2);
plot(f,fasehoek,'c');

asse=axis;
axis([min_f max_f asse(3) asse(4)]);
xlabel(['Frekwensie [Hz]  (TX-fs=' sprintf('%4.2f',TX_fs/1000) ' kHz)']);
ylabel('Fasehoek [ ]');
zoom on;

% -------------- Vertoon die anti-oorvleuelingsfilter se frekwensiegedrag (fig 2) ---------------

spektrum1=20*log10(abs(freqs(-TX_teller,TX_noemer,2*pi*f)));
fasehoek1=unwrap(angle(freqs(-TX_teller,TX_noemer,2*pi*f)))*180/pi;

figure(2);
clf;

subplot(2,1,1);
hold on;

plot(f,spektrum1,'c');

if alfa == 0						% vir die grensgeval van alfa=0 slegs reguit lyne
  line([min_f fb/2], [0 0]);
  line([fb/2 fb/2], [0 -100]);
else
  plot(f,20*log10(ideale_spektrum), 'y:');
end;

hold off;
axis([min_f max_f min_dB max_dB]);
title(['Aanwins- en faseweergawe van poetsfilter (geel: alfa=' sprintf('%3.2f',alfa) ')']);
ylabel('Aanwins [dB]');

subplot(2,1,2);
plot(f,fasehoek1,'c');

asse=axis;
axis([min_f max_f asse(3) 0]);
xlabel(['Frekwensie [Hz]  (TX-fs=' sprintf('%4.2f',TX_fs/1000) ' kHz)']);
ylabel('Fasehoek [ ]');
zoom on;

% ----------------- Vertoon die ontvangerfilter se frekwensiegedrag (fig 3) ---------------------

spektrum2=20*log10(abs(freqs(RX_teller,RX_noemer,2*pi*f)));
fasehoek2=unwrap(angle(freqs(RX_teller,RX_noemer,2*pi*f)))*180/pi;

figure(3);
clf;

subplot(2,1,1);
hold on;

plot(f,spektrum2,'c');

if alfa == 0						% vir die grensgeval van alfa=0 slegs reguit lyne
  line([min_f fb/2], [0 0]);
  line([fb/2 fb/2], [0 -100]);
else
  plot(f,20*log10(ideale_spektrum), 'y:');
end;
hold off;

axis([min_f max_f min_dB max_dB]);
title(['Aanwins- en faseweergawe van ontvangerfilter (geel: alfa=' sprintf('%3.2f',alfa) ')']);
ylabel('Aanwins [dB]');

subplot(2,1,2);
plot(f,fasehoek2,'c');

asse=axis;
axis([min_f max_f asse(3) 0]);
xlabel(['Frekwensie [Hz]  (TX-fs=' sprintf('%4.2f',TX_fs/1000) ' kHz)']);
ylabel('Fasehoek [ ]');
zoom on;

% ------------ Vertoon die impulsresponse by die uittrees van die filters (fig 4) ---------------

figure(4);
clf;

bisperiodes=linspace(0,length(impuls)*fb/TX_fs-length(impuls)*fb/TX_fs/length(impuls),length(impuls));

stairs(bisperiodes*TX_fs/RX_fs, impuls);
hold on;

t=linspace(0,length(impuls)/RX_fs,length(impuls));
impuls=-lsim(TX_teller, TX_noemer, impuls,t);
plot(bisperiodes*TX_fs/RX_fs, impuls,'c:');

impuls=lsim(RX_teller, RX_noemer, impuls,t);
plot(bisperiodes*TX_fs/RX_fs, impuls,'b--');

axis([0 length(impuls)*fb/RX_fs-length(impuls)*fb/RX_fs/length(impuls) -0.4 1.2]);
title('Impulsrespons n transversale filter, poetsfilter en ontvangerfilter');
xlabel('Bisperiodes');
ylabel('Genormaliseerde spanning [V]');
zoom on;

% ------------------------- Vertoon seine by modem-toetspunte (fig 5) ---------------------------

fs_monster=ones(TX_fs/fb, 1);				% monster die geskommelde ry teen TX-fs
TX_monster=fs_monster*TX_skommel;
TX_monster=TX_monster(:)';

figure(5);
clf;
hold on;

bisperiodes=linspace(0,length(TX_oudio)*fb/RX_fs,length(TX_monster)+bisse*TX_fs/fb);
stairs(bisperiodes,[zeros(1, bisse*TX_fs/(2*fb)) TX_monster zeros(1, bisse*TX_fs/(2*fb))]+15);

bisperiodes=linspace(0,length(TX_oudio)*fb/RX_fs,length(TX_oudio));
stairs(bisperiodes,TX_zoh+13.5);
plot(bisperiodes,TX_oudio+11.5);

bisperiodes=linspace(0,length(TX_oudio)*fb/RX_fs,length(RX_oudio)-(RX_fs/TX_fs-1));
plot(bisperiodes,RX_oudio(1:length(RX_oudio)-(RX_fs/TX_fs-1))+9.5);
plot(bisperiodes,RX_oog(1:length(RX_oudio)-(RX_fs/TX_fs-1))+7.5);
stairs(bisperiodes,RX_dig(1:length(RX_oudio)-(RX_fs/TX_fs-1))+5);
stairs(bisperiodes,RX_proto(1:length(RX_oudio)-(RX_fs/TX_fs-1))+3);
stairs(bisperiodes,RX_klok(1:length(RX_oudio)-(RX_fs/TX_fs-1))+1);

axis([0 length(RX_oudio)*fb/RX_fs 0 17]);
title('TX: skommel, zoh, oudio, RX: oudio, oog, dig, proto, klok');
xlabel('Bisperiodes');
ylabel('Genormaliseerde spanning [V]');
hold off;
zoom on;

% --------------------------- Vertoon ontvanger-oogpatroon (fig 6) ------------------------------

begin=576;						% afset sodat oog simmetries binne raam gespasieer is

for i=1:length(TX_skommel)-oge,				% kap sein op in veelvoude van bislengtes volgens die aantal o wat vertoon moet word
  oog(i,1:oge*RX_fs/fb+1)=RX_oog(begin+(i-1)*RX_fs/fb:begin+(i-1)*RX_fs/fb+oge*RX_fs/fb)+0.5;
end;	

figure(6);
clf;
hold on;

bisperiodes=linspace(0,oge,oge*RX_fs/fb+1);
for i=1:size(oog,1),
  plot(bisperiodes,oog(i,1:length(oog(1,:))));
end;

axis([0 oge -0.5 1.5]);
title('Oogkarakteristiek van modem');
xlabel('Bisperiode(s)');
ylabel('Genormaliseerde spanning [V]');
hold off;
zoom on;

% ------------- Vertoon oombliklike klokspeling tussen RX-klok en Ideale klok (fig 7) -----------

for i=1:size(oog,1),					% bereken maksimum toelaatbare afset tussen RX-klok en Ideale klok volgens oogkarakteristiek
  for j=1:RX_fs/(2*fb)
    if xor(0.5*sign(oog(i,j)-0.5)+0.5, 0.5*sign(oog(i,j+1)-0.5)+0.5)==1
      grens(i)=RX_fs/(2*fb)-j+1;
    elseif xor(0.5*sign(oog(i,j)-0.5)+0.5, 0.5*sign(oog(i,1)-0.5)+0.5)==0
      grens(i)=RX_fs/(2*fb);
    end;
  end;
end;
maks_speling=min(grens);

figure(7);
clf;
hold on;

bisperiodes=linspace(0,length(afset_rekord),length(afset_rekord));
stem(bisperiodes,afset_rekord);

punte=linspace(0,length(afset_rekord),2);
plot(punte, maks_speling*[1 1], 'r--');
plot(punte, -maks_speling*[1 1], 'r--');

axis([0 length(afset_rekord) -RX_fs/(2*fb) RX_fs/(2*fb)]);
title('Speling tussen RX-klok en Ideale-klok (binne grense: 100% akkuraat)');
xlabel('Bisperiodes');
ylabel('Monsterpunte oor periode');
hold off;
zoom on;

% ------ Vertoon DCD-kromme (aanduiding van sluiting tussen RX-klok en RX-proto) (fig 8) --------

min_drempel=((RX_fs/(2*fb)-maks_speling)*fb/RX_fs)+0.25;	% skakel die maksimum oombliklike klokspeling om na 'n genormaliseerde spanning

figure(8);
clf;
hold on;

bisperiodes=linspace(0,length(TX_oudio)*fb/RX_fs,length(RX_oudio)-(RX_fs/TX_fs-1));
plot(bisperiodes,sluit_rekord(1:length(RX_oudio)-(RX_fs/TX_fs-1)));

punte=linspace(0,length(TX_oudio)*fb/TX_fs,2);
plot(punte, drempel*[1 1], 'b:');
plot(punte, min_drempel*[1 1], 'r--');

axis([0 length(RX_oudio)*fb/RX_fs 0 0.8]);
title('Sluiting tussen RX-klok en RX-proto (b blou drempel: DCD geldig)');
xlabel('Bisperiodes');
ylabel('Genormaliseerde spanning [V]');
hold off;
zoom on;

% --------------- Identifiseer datamerker binne RX-data en dekodeer indien gevind ---------------

datawyser=[];
for i=1:length(RX_data)-8*length(merker)+1						% soek datamerker
  if RX_data(i:i+8*length(merker)-1)==tex2bin(merker)
    datawyser=i+length(tex2bin(merker));
  end;
end;

duur=cputime-begintyd
clc;

if isempty(datawyser)==0 & (length(RX_data)-datawyser+1)>=length(tex2bin(boodskap))	% indien datamerker gevind en hele boodskap ontvang
  RX_boodskap=RX_data(datawyser:datawyser+8*length(boodskap)-1);			% onttrek boodskap
  verkeerd=xor(RX_boodskap,tex2bin(boodskap));						% tel foutiewe bisse
  korreksie=DCD_rekord(datawyser:datawyser+8*length(boodskap)-1);			% tel hoeveel van hulle nie ongesiens sou deurglip nie
  disp([' ']);
  disp(['Gestuur:       ' boodskap]);
  disp(['Ontvang:       ' bin2tex(RX_boodskap)]);
  disp([' ']);
  disp(['Bisse gestuur: ' num2str(8*length(boodskap))]);
  disp(['Uitgefilter:   ' num2str(sum(~korreksie))]);
  disp(['Foute:         ' num2str(sum(korreksie.*verkeerd))]);
  disp([' ']);
else											% indien nie, dui dit aan
  disp([' ']);
  disp('Die merker kon nie gevind word nie.');
  disp([' ']);
end;

figure(7);
figure(6);
figure(5);
figure(4);
figure(3);
figure(2);
figure(1);